# Copyright (c) Shopkick 2017
# See LICENSE for details.
'''
Shell
-----
Running subprocesses with a shell-like interface.
'''
import contextlib
import os
import tempfile
import time
import signal
import subprocess

import attr

class ProcessError(Exception):

    """
    A process has exited with non-zero status.
    """

    def __init__(self, *args):
        super(ProcessError, self).__init__()
        self._args = args
        self.returncode = args[0]
        if len(self._args) > 1:
            self.output = args[1]
        if len(self._args) > 2:
            self.error = args[1]

    def __repr__(self):
        return 'ProcessError{}'.format(repr(self._args))

    __str__ = __repr__

    def __getitem__(self, i):
        return self._args[i]

    def __iter__(self):
        return iter(self._args) # pragma: no cover

@attr.s
class Shell(object):

    """
    Run subprocesses.

    Init arguments:

    :param cwd: current working directory (default is process's current working directory)
    :param env: environment variables dict (default is a copy of the process's environment)
    """

    _procs = attr.ib(init=False, default=attr.Factory(list))

    _cwd = attr.ib(init=False, default=attr.Factory(os.getcwd))

    _env = attr.ib(init=False, default=attr.Factory(lambda: dict(os.environ)))

    def redirect(self, command, outfp, errfp, cwd=None):
        """
        Run a process, while its standard error and output go to pre-existing files

        :param command: list of arguments
        :param outfp: output file object
        :param errfp: error file object
        :param cwd: current working directory (default is to use the internal working directory)
        :raises: :code:`ProcessError` with return code
        """
        proc = self.popen(command, stdin=subprocess.PIPE, stdout=outfp, stderr=errfp, cwd=cwd)
        proc.communicate('')
        retcode = proc.wait()
        if retcode != 0:
            raise ProcessError(retcode)

    def batch(self, command, cwd=None):
        """
        Run a process, wait until it ends and return the output and error

        :param command: list of arguments
        :param cwd: current working directory (default is to use the internal working directory)
        :returns: pair of standard output, standard error
        :raises: :code:`ProcessError` with (return code, standard output, standard error)
        """
        with tempfile.NamedTemporaryFile() as stdout, \
             tempfile.NamedTemporaryFile() as stderr:
            proc = self.popen(command, stdin=subprocess.PIPE, stdout=stdout, stderr=stderr, cwd=cwd)
            proc.communicate('')
            retcode = proc.wait()
            self._procs.remove(proc)
            stdout.seek(0)
            stderr.seek(0)
            stdout_contents = stdout.read()
            stderr_contents = stderr.read()
            ## Log contents of stdout, stderr
            if retcode != 0:
                raise ProcessError(retcode, stdout_contents, stderr_contents)
            else:
                return stdout_contents, stderr_contents

    def interactive(self, command, cwd=None):
        """
        Run a process, while its standard output and error go directly to ours.

        :param command: list of arguments
        :param cwd: current working directory (default is to use the internal working directory)
        :raises: :code:`ProcessError` with (return code, standard output, standard error)
        """
        proc = self.popen(command, cwd=cwd)
        retcode = proc.wait()
        self._procs.remove(proc)
        if retcode != 0:
            raise ProcessError(retcode)

    def popen(self, command, **kwargs):
        """
        Run a process, giving direct access to the :code:`subprocess.Popen` arguments.

        :param command: list of arguments
        :param kwargs: keyword arguments passed to :code:`subprocess.Popen`
        :returns: a :code:`Process`
        """
        if kwargs.get('cwd') is None:
            kwargs['cwd'] = self._cwd
        if kwargs.get('env') is None:
            kwargs['env'] = self._env
        proc = subprocess.Popen(command, **kwargs)
        self._procs.append(proc)
        return proc

    def setenv(self, key, val):
        """
        Set internal environment variable.

        Changes internal environment in which subprocesses will be run.
        Does not change the process's own environment.

        :param key: name of variable
        :param value: value of variable
        """
        if val is None:
            if key in self._env:
                del self._env[key]
            return
        key = str(key)  # keys must be strings
        val = str(val)  # vals must be strings
        self._env[key] = val

    def getenv(self, key):
        """
        Get internal environment variable.

        Return value of variable in internal  environment in which subprocesses will be run.

        :param key: name of variable
        :returns: value of variable
        :raises: :code:`KeyError` if key is not in environment
        """
        return self._env[key]

    def chdir(self, path):
        """
        Change internal current working directory.

        Changes internal directory in which subprocesses will be run.
        Does not change the process's own current working directory.

        :param path: new working directory
        """
        self._cwd = os.path.join(self._cwd, path)

    def reap_all(self):
        """
        Kill, as gently as possible, all processes.

        Loop through all processes and try to kill them with
        a sequence of :code:`SIGINT`, :code:`SIGTERM` and
        :code:`SIGKILL`.
        """
        for proc in self._procs:
            ret_code = proc.poll()
            if ret_code is None:
                proc.send_signal(signal.SIGINT)
                time.sleep(3)
            ret_code = ret_code or proc.poll()
            if ret_code is None: # pragma: no coverage
                proc.terminate()
                time.sleep(3)
            ret_code = ret_code or proc.poll() # pragma: no coverage
            if ret_code is None: # pragma: no coverage
                proc.kill()

    def clone(self):
        """
        Clone the shell object.

        :returns: a new Shell object with a copy of the environment dictionary
        """
        return attr.assoc(self, _env=dict(self._env), _procs=[])

@contextlib.contextmanager
def autoexit_code():
    """
    Context manager that translates :code:`ProcessError` to immediate process exit.
    """
    try:
        yield
    except ProcessError as pexc:
        raise SystemExit(pexc[0])



# Copyright (c) Shopkick 2017
# See LICENSE for details.
"""
Executor
--------

Construct command-line lists.

:const:`NO_VALUE` -- indicate an option with no value (a boolean option)
"""
import functools

import six

import singledispatch

import attr

NO_VALUE = object()

@attr.s(frozen=True)
class Eq(object):
    """Wrap a string to indicate = option

    Wrap a string to indicate that the option
    *has* to be given as '--name=value'
    rather than the usually equivalent and
    more automation-friendly '--name value'

    :code:`git show --format`, I'm looking
    at you.
    """

    content = attr.ib()

@singledispatch.singledispatch
def _keyword_arguments(_value, key):
    yield key

@_keyword_arguments.register(Eq)
def _keyword_arguments_eq(value, key):
    yield key + '=' + value.content

@_keyword_arguments.register(str)
def _keyword_arguments_str(value, key):
    yield key
    yield value

@_keyword_arguments.register(int)
def _keyword_arguments_int(value, key):
    yield key
    yield str(value)

@_keyword_arguments.register(list)
def _keyword_arguments_list(value, key):
    for thing in value:
        yield key
        yield thing

@_keyword_arguments.register(dict)
def _keyword_arguments_dict(value, key):
    for in_k, thing in value.items():
        yield key
        yield '{}={}'.format(in_k, thing)

def cmd(binary, subcommand, *args, **kwargs):
    """
    Construct a command line for a "modern UNIX" command.

    Modern UNIX command do a closely-related-set-of-things and do it well.
    Examples include :code:`apt-get` or :code:`git`.

    :param binary: the name of the command
    :param subcommand: the subcommand used
    :param args: positional arguments (put last)
    :param kwargs: options
    :returns: list of arguments that is suitable to be passed to :code:`subprocess.Popen`
              and friends.

    When specifying options, the following assumptions are made:

    * Option names begin with :code:`--` and any :code:`_` is assumed to be a :code:`-`
    * If the value is :code:`NO_VALUE`, this is a "naked" option.
    * If the value is a string or an int, these are presented as the value of the option.
    * If the value is a list, the option will be repeated multiple times.
    * If the value is a dict, the option will be repeated multiple times, and
      its values will be :code:`<KEY>=<VALUE>`.
    """
    ret = [binary, subcommand]
    for key, value in kwargs.items():
        key = '--' + key.replace('_', '-')
        ret.extend(_keyword_arguments(value, key))
    ret.extend(args)
    return ret

@attr.s(frozen=True)
class _PreparedCommand(object):

    _cmd = attr.ib()
    _shell = attr.ib()

    def batch(self, *args, **kwargs):
        """Run the shell's batch"""
        return self._shell.batch(self._cmd, *args, **kwargs)

    def interactive(self, *args, **kwargs):
        """Run the shell's interactive"""
        return self._shell.interactive(self._cmd, *args, **kwargs)

    def redirect(self, *args, **kwargs):
        """Run the shell's redirect"""
        return self._shell.redirect(self._cmd, *args, **kwargs)

    def popen(self, *args, **kwargs):
        """Run the shell's popen"""
        return self._shell.popen(self._cmd, *args, **kwargs)


@attr.s(frozen=True)
class Command(object):

    """
    A command is something that can be bound to an executor.
    Commands get automatically bound if defined as members of an executor.

    :param name: the name of a 'Modern UNIX' command (i.e., something with subcommands).
    """
    _name = attr.ib()

    def bind(self, executor, _dummy=None):
        """
        Bind a command to an executor.

        :param executor: the executor to bind to
        :returns: something that has methods :code:`batch`, :code:`interactive` and :code:`popen`
                  methods.
        """
        return _ExecutoredCommand(executor, self._name)

    __get__ = bind

@attr.s(frozen=True)
class _ExecutoredCommand(object):

    _executor = attr.ib()

    _name = attr.ib()

    def __call__(self, *args, **kwargs):
        return self._executor.prepare(self._name, *args, **kwargs)

    def __getattr__(self, subcommand):
        subcommand = subcommand.rstrip('_').replace('_', '-')
        return functools.partial(self._executor.prepare, self._name, subcommand)

@attr.s(frozen=True)
class Executor(object):

    """
    Executes commands.

    Init parameters:

    :param shell: something that actually runs subprocesses.
                  Should match the interface of :code:`Shell`.
    :param pypi: optional. An extra index URL.
    :param commands: optional. An iterable of strings which are commands to suppport.

    The default commands that are supported are :code:`git`, :code:`pip`, :code:`conda`,
    :code:`docker`, :code:`docker_machine`.
    """

    _shell = attr.ib()
    _pypi = attr.ib(default=None)
    _commands = attr.ib(default=attr.Factory(set), converter=set)

    git = Command('git')
    pip = Command('pip')
    conda = Command('conda')
    docker = Command('docker')
    docker_machine = Command('docker-machine')

    def __getattr__(self, name):
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError(name) # Reserved Python names not supported as commands
        if name not in self._commands:
            raise AttributeError(name)
        name = name.replace('_', '-')
        return Command(name).bind(self)

    def add_command(self, name):
        """
        Add a new command.

        :param name: name of command
        """
        self._commands.add(name)

    def prepare(self, command, subcommand, *args, **kwargs):
        """
        Prepare a command (inspired by SQL statement preparation).

        :param command: name of command (e.g., :code:`apt-get`)
        :param subcommand: name of sub-command (e.g., :code:`install`)
        :param args: positional arguments
        :param kwargs: option arguments
        :returns: something that supports batch/interactive/popen
        """
        return _PreparedCommand(cmd=cmd(command, subcommand, *args, **kwargs),
                                shell=self._shell.clone())

    def command(self, args):
        """
        Prepare a command from a raw argument list.

        :param args: argument list
        :returns: something that supports batch/interactive/popen
        """
        return _PreparedCommand(args, shell=self._shell.clone())

    def in_docker_machine(self, machine):
        """
        Return an executor where all docker commands would point at a specific Docker machine.

        :param machine: name of machine
        :returns: a new executor
        """
        new_shell = self._shell.clone()
        output, _ignored = self.docker_machine.env(machine, shell='cmd').batch()
        for line in output.splitlines():
            directive, args = line.split(None, 1)
            if directive != 'SET':
                continue
            key, value = args.split('=', 1)
            new_shell.setenv(key, value)
        return attr.assoc(self, _shell=new_shell)

    def patch_env(self, **kwargs):
        """
        Return a new executor where the environment is patched with the given attributes

        :param kwargs: new environment variables
        :returns: new executor with a shell with a patched environment.
        """
        new_shell = self._shell.clone()
        for key, value in kwargs.items():
            new_shell.setenv(key, value)
        return attr.evolve(self, shell=new_shell)

    def chdir(self, path):
        """
        Return a new executor where the working directory is different.

        :param path: new path
        :returns: new executor with a different working directory
        """
        new_shell = self._shell.clone()
        new_shell.chdir(path)
        return attr.evolve(self, shell=new_shell)

    def in_virtualenv(self, envpath):
        """
        Return an executor where all Python commands would point at a specific virtual environment.

        :param envpath: path to virtual environment
        :returns: a new executor
        """
        new_shell = self._shell.clone()
        new_shell.setenv('VIRTUAL_ENV', envpath)
        new_shell.setenv('PYTHONHOME', None)
        try:
            old_path = new_shell.getenv('PATH')
            new_path = envpath + '/bin' + ':' + old_path
        except KeyError:
            new_path = envpath + '/bin'
        new_shell.setenv('PATH', new_path)
        return attr.assoc(self, _shell=new_shell)

    def pip_install(self, pkg_ids, index_url=None):
        """
        Use pip to install packages

        :param pkg_ids: an list of package names
        :param index_url: (optional) an extra PyPI-compatible index
        :raises: :code:`ProcessError` if the installation fails
        """
        if index_url is None:
            index_url = self._pypi
        if index_url is not None:
            trusted_host = six.moves.urllib.parse.urlparse(index_url).netloc
            kwargs = dict(extra_index_url=index_url, trusted_host=trusted_host)
        else:
            kwargs = {}
        mycmd = self.pip.install(*pkg_ids, **kwargs)
        return mycmd.batch()

    def conda_install(self, pkg_ids, channels=None):
        """
        Use conda to install packages

        :param pkg_ids: an list of package names
        :param channels: (optional) a list of channels to install from
        :raises: :code:`ProcessError` if the installation fails
        """
        mycmd = self.conda.install(quiet=NO_VALUE, yes=NO_VALUE, show_channel_urls=NO_VALUE,
                                   channel=(channels or []), *pkg_ids)
        return mycmd.batch()



# Copyright (c) Contributors 2020
# See License for details
"""
Provides seashore version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update seashore` to change this file.

from incremental import Version

__version__ = Version('seashore', 20, 6, 0)
__all__ = ["__version__"]



# Copyright (c) Shopkick 2017
# See LICENSE for details.
"""
Seashore
=======

Seashore is a collection of shell abstractions.
"""
from seashore.executor import Executor, NO_VALUE, Eq
from seashore.shell import Shell, ProcessError
from seashore._version import __version__

__all__ = ['Executor', 'NO_VALUE', 'Eq', 'Shell', 'ProcessError', '__version__']



# Copyright (c) Shopkick 2017
# See LICENSE for details.
"""Tests for seashore.shell"""

import sys
import tempfile
import unittest
import traceback

from seashore import shell

class ShellTest(unittest.TestCase):

    """Tests for Shell()"""

    def setUp(self):
        """create a new shell object"""
        self.shell = shell.Shell()

    def test_redirect(self):
        """redirect sends output to temp file"""
        python_script = "import sys;sys.stdout.write('hello');sys.stderr.write('goodbye')"
        with tempfile.NamedTemporaryFile() as stdout, \
             tempfile.NamedTemporaryFile() as stderr:
            self.shell.redirect([sys.executable, '-c', python_script], stdout, stderr)
            stdout.seek(0)
            out = stdout.read()
            stderr.seek(0)
            err = stderr.read()
        self.assertEqual(out, b'hello')
        self.assertEqual(err, b'goodbye')

    def test_failed_redirect(self):
        """redirect sends output to temp file and raises if it fails"""
        python_script = ("import sys;sys.stdout.write('hello');sys.stderr.write('goodbye');"
                         "sys.exit(1)")
        with tempfile.NamedTemporaryFile() as stdout, \
             tempfile.NamedTemporaryFile() as stderr:
            with self.assertRaises(shell.ProcessError):
                self.shell.redirect([sys.executable, '-c', python_script], stdout, stderr)
            stdout.seek(0)
            out = stdout.read()
            stderr.seek(0)
            err = stderr.read()
        self.assertEqual(out, b'hello')
        self.assertEqual(err, b'goodbye')

    def test_batch(self):
        """batch mode returns contents of stdout/stderr from subprocesses"""
        python_script = "import sys;sys.stdout.write('hello');sys.stderr.write('goodbye')"
        out, err = self.shell.batch([sys.executable, '-c', python_script])
        self.assertEqual(out, b'hello')
        self.assertEqual(err, b'goodbye')

    def test_failed_batch(self):
        """processes exiting with non-zero code causes an exception in batch mode"""
        python_script = b"raise SystemExit(1)"
        with self.assertRaises(shell.ProcessError):
            self.shell.batch([sys.executable, '-c', python_script])

    def test_failed_interactive(self):
        """processes exiting with non-zero code causes an exception in interactive mode"""
        python_script = "raise SystemExit(1)"
        with self.assertRaises(shell.ProcessError):
            self.shell.interactive([sys.executable, '-c', python_script])

    def test_env(self):
        """setting an environment variable in the shell propagates to subprocesses"""
        self.shell.setenv('SPECIAL', 'emett')
        python_script = b'import sys,os;sys.stdout.write(os.environ["SPECIAL"])'
        out, _ignored = self.shell.batch([sys.executable, '-c', python_script])
        self.assertEqual(out, b'emett')
        self.assertEqual(self.shell.getenv('SPECIAL'), 'emett')

    def test_chdir(self):
        """changing the shell's directory effects the cwd of processes it runs"""
        self.shell.chdir('/')
        python_script = b'import sys,os;sys.stdout.write(os.getcwd())'
        out, _ignored = self.shell.batch([sys.executable, b'-c', python_script])
        self.assertEqual(out, b'/')

    def test_reaper(self):
        """killing a process terminates it with a negative signal"""
        python_script = b'import time;time.sleep(100000)'
        proc = self.shell.popen([sys.executable, b'-c', python_script])
        self.shell.reap_all()
        self.assertLess(proc.wait(), 0)

    def test_clone(self):
        """cloned shell's environment does not interfere with the original"""
        new_shell = self.shell.clone()
        new_shell.setenv('SPECIAL', 'lucy')
        self.shell.setenv('SPECIAL', 'emett')
        python_script = b'import sys,os;sys.stdout.write(os.environ["SPECIAL"])'
        out, _ignored = new_shell.batch([sys.executable, '-c', python_script])
        self.assertEqual(out, b'lucy')

    def test_env_none(self):
        """passing env variable as none deletes it"""
        self.shell.setenv('SPECIAL', 'lucy')
        self.shell.setenv('SPECIAL', None)
        python_script = b'import sys,os;sys.stdout.write(os.environ.get("SPECIAL", "emett"))'
        out, _ignored = self.shell.batch([sys.executable, '-c', python_script])
        self.assertEqual(out, b'emett')

class ProcessErrorTest(unittest.TestCase):

    """Tests that check process error is useful"""

    def test_returncode(self):
        """returncode attribtue is set to passed-in value"""
        self.assertEqual(shell.ProcessError(13).returncode, 13)

    def test_output(self):
        """output attribtue is set to passed-in value"""
        self.assertEqual(shell.ProcessError(13, "woo").output, "woo")

    def test_repr(self):
        """repr contains output, error and code"""
        procerr = shell.ProcessError(13, "myout42", "myerr42")
        stringified = repr(procerr)
        self.assertIn('13', stringified)
        self.assertIn('myout42', stringified)
        self.assertIn('myerr42', stringified)

    def test_trace(self):
        """traceback has args"""
        try:
            raise shell.ProcessError(13, "myout8080", "myerr8080")
        except shell.ProcessError:
            trace = traceback.format_exc()
        stringified = trace.splitlines()[-2]
        self.assertIn('13', stringified)
        self.assertIn('myout8080', stringified)
        self.assertIn('myerr8080', stringified)


class AutoexitTest(unittest.TestCase):

    """Tests for autoexit_code"""

    def test_success(self):
        """autoexit_code wrapping success does not exit"""
        with shell.autoexit_code():
            pass
        self.addCleanup(lambda: None)

    def test_failure(self):
        """autoexit_code wrapping a ProcessError raise exits"""
        with self.assertRaises(SystemExit):
            with shell.autoexit_code():
                raise shell.ProcessError(13)



# Copyright (c) Shopkick 2017
# See LICENSE for details.
from __future__ import absolute_import, division, print_function

import setuptools

with open('README.rst') as fp:
    long_description = fp.read()

setuptools.setup(
    name='seashore',
    maintainer='Shopkick',
    maintainer_email='dev@shopkick.com',
    url="http://github.com/shopkick/seashore",
    use_incremental=True,
    setup_requires=['incremental'],
    license='MIT',
    install_requires=[
        'incremental',
        'attrs>=19.3.0',
        'singledispatch',
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages('src'),
    description='A collection of shell abstractions',
    long_description=long_description,
)



# Copyright (c) Shopkick 2017
# See LICENSE for details.
# pragma pylint: disable=too-many-boolean-expressions
# pragma pylint: disable=too-many-return-statements
# pragma pylint: disable=too-many-branches
# pragma pylint: disable=too-many-public-methods
# The pragmas above are to deal with the DummyShell
# class. It should probably be broken up and define a class
# per-test. While defining a class in a function is usually
# a bad idea, because of memory leaks, in test functions
# it should probably be ok.
# See issue #21
"""Test seashore.executor"""

import os
import unittest

import attr

from seashore import executor

@attr.s
class DummyShell(object):

    """Dummy Shell"""

    _env = attr.ib(default=attr.Factory(dict))

    _cwd = attr.ib(default="")

    def clone(self):
        """Return a copy of the shell"""
        return attr.evolve(self, env=dict(self._env))

    def setenv(self, key, value):
        """Set an environment variable"""
        self._env[key] = value

    def chdir(self, path):
        """change directory"""
        self._cwd = os.path.join(self._cwd, path)

    def getenv(self, key):
        """Get an environment variable"""
        return self._env[key]

    def batch(self, *args, **kwargs):
        """(Pretend to) run a command in batch mode"""
        if args == ('git rev-parse HEAD'.split(),) and kwargs == {}:
            return '777', ''
        if args == ('docker-machine env --shell cmd confluent'.split(),) and kwargs == {}:
            return ('SET DOCKER_TLS_VERIFY=1\n'
                    'SET DOCKER_HOST=tcp://192.168.99.103:2376\n'
                    'SET DOCKER_CERT_PATH=/Users/u/.docker/machine/machines/confluent\n'
                    'SET DOCKER_MACHINE_NAME=confluent\n'
                    'REM Run this command to configure your shell: \n'
                    'REM 	@FOR /f "tokens=*" %i IN (\'docker-machine env --shell cmd confluent\') '
                    'DO @%i\n', '')
        if (len(args) == 1 and args[0][:2] == 'docker run'.split() and
                set(args[0][2:-1]) == set('--interactive --remove --terminal'.split()) and
                args[0][-1] == 'a-machine:a-tag' and
                self._env['DOCKER_MACHINE_NAME'] == 'confluent' and
                self._env['DOCKER_CERT_PATH'] == '/Users/u/.docker/machine/machines/confluent' and
                self._env['DOCKER_TLS_VERIFY'] == '1' and
                self._env['DOCKER_HOST'] == 'tcp://192.168.99.103:2376'):
            return 'hello\r\n', ''
        if args == ('pip install attrs'.split(),):
            return 'attrs installed', ''
        if (args == ('pip install a-local-package'.split(),) and
                self._env['VIRTUAL_ENV'] == '/appenv'):
            return 'a-local-package installed', ''
        if args == ('apt-get update'.split(),):
            return 'update finished successfully', ''
        if args == ('echo hello'.split(),):
            return 'hello\n', ''
        if (len(args) == 1 and args[0][:2] == 'pip install'.split() and
                args[0][-1] == 'attrs' and
                '--trusted-host' in args[0] and
                args[0][args[0].index('--trusted-host')+1] == 'orbifold.xyz' and
                '--extra-index-url' in args[0] and
                args[0][args[0].index('--extra-index-url')+1] == 'http://orbifold.xyz'):
            return 'attrs installed from orbifold', ''
        if (len(args) == 1 and args[0][:2] == 'conda install'.split() and
                set(args[0][2:-1]) == set('--show-channel-urls --quiet --yes'.split()) and
                args[0][-1] == 'numpy'):
            return 'numpy installed', ''
        if (len(args) == 1 and args[0][:2] == 'docker run'.split() and
                args[0][2] == '--env' and
                args[0][4] == '--env' and
                set([args[0][3], args[0][5]]) == set(['SONG=awesome', 'SPECIAL=emett']) and
                args[0][-1] == 'lego:1'):
            return 'everything', ''
        if args == ('do-stuff special --verbosity 5'.split(),):
            return 'doing stuff very specially', ''
        if args == ('do-stuff special --verbose'.split(),):
            return 'doing stuff slightly more verbosely', ''
        if (len(args) == 1 and args[0][:2] == 'chat mention'.split() and
                args[0][2] == '--person' and
                args[0][4] == '--person' and
                set([args[0][3], args[0][5]]) == set(['emett', 'lucy'])):
            return 'mentioning folks', ''
        if args == (['pwd'],):
            return self._cwd, ''
        if args == ('docker exec 3433 echo yay'.split(),):
            return 'yay\r\n', ''
        if (args[0][:2] == 'git show'.split() and
                '--no-patch' in args[0] and
                '--format=%ct' in args[0] and
                len(args[0]) == 4):
            return '1496798292', ''
        raise ValueError(self, args, kwargs)

    def interactive(self, *args, **kwargs):
        """(Pretend to) run a command in interactive mode"""
        if args == (['python'],):
            return
        raise ValueError(self, args, kwargs)

    def popen(self, *args, **kwargs):
        """(Pretend to) run a command in popen mode"""
        if args == (['grep', 'foo'],):
            return
        raise ValueError(self, args, kwargs)

    def redirect(self, cmd, outfp, errfp, *args, **kwargs):
        """(Pretend to) redirect a command"""
        if cmd == ['docker', 'run', 'confluent'] and outfp == 1 and errfp == 2:
            return
        raise ValueError(self, cmd, outfp, errfp, args, kwargs)


class ExecutorTest(unittest.TestCase):

    """Test Executor"""

    def setUp(self):
        """build an executor with a dummy shell"""
        self.shell = DummyShell()
        self.executor = executor.Executor(self.shell)

    def test_eq(self):
        """forcing eq yields equals-style option"""
        out, _err = self.executor.git.show(no_patch=None, format=executor.Eq('%ct')).batch()
        self.assertEqual(out, '1496798292')

    def test_redirect(self):
        """redirecting output and error"""
        self.executor.docker.run('confluent').redirect(outfp=1, errfp=2)

    def test_underscore(self):
        """underscore is converted to a dash"""
        out, err = self.executor.git.rev_parse('HEAD').batch()
        self.assertEqual(out, '777')
        self.assertEqual(err, '')

    def test_in_docker_machine(self):
        """calling in_docker_machine returns an executor that runs docker pointed at the machine"""
        new_executor = self.executor.in_docker_machine('confluent')
        output, _err = new_executor.docker.run('a-machine:a-tag', remove=executor.NO_VALUE,
                                               interactive=executor.NO_VALUE,
                                               terminal=executor.NO_VALUE).batch()
        self.assertEqual(output, 'hello\r\n')

    def test_in_virtualenv(self):
        """calling in_virtualenv returns an executor that runs pip in a virtual env"""
        new_executor = self.executor.in_virtualenv('/appenv')
        output, _err = new_executor.pip.install('a-local-package').batch()
        self.assertEqual(output, 'a-local-package installed')
        new_executor_one = self.executor.patch_env(PATH='/bin')
        new_executor_two = new_executor_one.in_virtualenv('/appenv')
        output, _err = new_executor_two.pip.install('a-local-package').batch()
        self.assertEqual(output, 'a-local-package installed')

    def test_call(self):
        """calling a built-in command runs it in the shell"""
        output, _error = self.executor.pip('install', 'attrs').batch()
        self.assertEqual(output, 'attrs installed')

    def test_arbitrary(self):
        """adding an arbitrary command allows access via attribute"""
        self.executor.add_command('apt_get')
        output, _error = self.executor.apt_get.update().batch()
        self.assertEqual(output, 'update finished successfully')

    def test_command(self):
        """using the command method passes the arguments directly to the shell"""
        output, _error = self.executor.command(['echo', 'hello']).batch()
        self.assertEqual(output, 'hello\n')

    def test_pip_install(self):
        """calling pip_install() runs 'pip install'"""
        output, _error = self.executor.pip_install(['attrs'])
        self.assertEqual(output, 'attrs installed')

    def test_pip_install_index(self):
        """passing the index_url param to pip_install passes it to the pip install command"""
        output, _error = self.executor.pip_install(['attrs'], index_url='http://orbifold.xyz')
        self.assertEqual(output, 'attrs installed from orbifold')

    def test_conda_install(self):
        """the conda_install() method calls 'conda install'"""
        output, _error = self.executor.conda_install(['numpy'])
        self.assertEqual(output, 'numpy installed')

    def test_interactive(self):
        """interactive mode does not fail"""
        self.executor.command(['python']).interactive()

    def test_non_existant_command(self):
        """non-existing commands raise AttributeError"""
        with self.assertRaises(AttributeError):
            self.executor.this_command_doesnt_exist.install().batch()

    def test_popen(self):
        """popen does not fail"""
        self.executor.command(['grep', 'foo']).popen()

    def test_dict_keywords(self):
        """prepare with dict value explodes the dict"""
        output, _err = self.executor.docker.run('lego:1', env=dict(SPECIAL='emett',
                                                                   SONG='awesome')).batch()
        self.assertEqual(output, 'everything')

    def test_none(self):
        """prepare with None option gives undecorated"""
        output, _err = self.executor.prepare('do-stuff', 'special', verbose=None).batch()
        self.assertEqual(output, 'doing stuff slightly more verbosely')

    def test_int(self):
        """prepare with int option stringifies the int"""
        output, _err = self.executor.prepare('do-stuff', 'special', verbosity=5).batch()
        self.assertEqual(output, 'doing stuff very specially')

    def test_list(self):
        """prepare with list option explodes into a list of options"""
        output, _err = self.executor.prepare('chat', 'mention', person=['emett', 'lucy']).batch()
        self.assertEqual(output, 'mentioning folks')

    def test_chdir(self):
        """changing directory changes the working directory"""
        new_executor = self.executor.chdir('foo/bar')
        output, _err = self.executor.command(['pwd']).batch()
        self.assertEqual(output, '')
        output, _err = new_executor.command(['pwd']).batch()
        self.assertEqual(output, 'foo/bar')

    def test_keyword(self):
        """using a trailing _ protects keywords"""
        output, _err = self.executor.docker.exec_('3433', 'echo', 'yay').batch()
        self.assertEqual(output, 'yay\r\n')
